@@ -138,7 +138,14 @@ BEGIN {
 				my ( $self, $data ) = @_;
 				my $new_data = $self->callback_and_reg( '.$reftype.' => $data );
 				if ( "'.uc($reftype).'" eq (reftype($new_data)||"") ) {
-					return $self->_register_mapping( $data, $self->SUPER::visit_'.$reftype.'( $new_data ) );
+					my $visited = $self->SUPER::visit_'.$reftype.'( $new_data );
+
+					no warnings "uninitialized";
+					if ( refaddr($visited) != refaddr($data) ) {
+						return $self->_register_mapping( $data, $visited );
+					} else {
+						return $visited;
+					}
 				} else {
 					return $self->_register_mapping( $data, $self->visit( $new_data ) );
 				}
@@ -165,15 +172,18 @@ sub callback_and_reg {
 	my $new_data = $self->callback( $name, $data, @args );
 
 	unless ( $self->ignore_return_values ) {
-		return $self->_register_mapping( $data, $new_data );
-	} else {
-		return $data;
+		no warnings 'uninitialized';
+		if ( refaddr($data) != refaddr($new_data) ) {
+			return $self->_register_mapping( $data, $new_data );
+		}
 	}
+
+	return $data;
 }
 
 sub visit_tied {
 	my ( $self, $tied, @args ) = @_;
-	$self->SUPER::visit_tied( $self->callback_and_reg( tied => $tied, @args ) );
+	$self->SUPER::visit_tied( $self->callback_and_reg( tied => $tied, @args ), @args );
 }
 
 __PACKAGE__;
@@ -290,6 +300,9 @@ C<object> callback.
 
 If the object C<isa> the class then the callback will fire.
 
+These callbacks are called from least derived to most derived by comparing the
+classes' C<isa> at construction time.
+
 =item object_no_class
 
 Called for every object that did not have a class callback.
